// By Josh Bevan 2020-02-08

// imports
import peasy.*;
import java.io.*; 
import java.util.*;

// initializations
Table table;
int numCols, numRows, earliestYear, latestYear, rangeYears;

PVector[] particles;

int[] classes;
int numClasses;
float[] errors;
String[] titles;
int[] publishingYears;

Book[] books;
Topic[] topics;
ArrayList<Year> years;
PShape theVisualization;
PeasyCam cam;

double x = 0;
double y = 0;
double z = 0;

// tweak zone
//
//
float innerCircleRadius = -20;
float distBetweenSlices = 2;
int textSize = 10;
color textColor = color(80, 70, 200);
int hueOffset = 50;
color background = color(20, 10, 50);
//
//
// tweak zone

void setup(){
  fullScreen(P3D);
  // set up camera
  cam = new PeasyCam(this, 100000);
  cam.setMinimumDistance(50);
  cam.setMaximumDistance(500);
  
  // set up miscellaneous stuff
  colorMode(HSB, 360, 100, 100);
  textSize(textSize);
  PFont mono = loadFont("MalgunGothicBold-10.vlw");
  textFont(mono);
  
  // pull the data out of the CSV
  table = loadTable("dimensionReducedBib8.csv");
  numCols = table.getColumnCount();
  numRows = table.getRowCount();
  
  particles = new PVector[numRows];
  titles = new String[numRows];
  classes = new int[numRows];
  errors = new float[numRows];
  publishingYears = new int[numRows];
  books = new Book[numRows];
  
  int i = 0;
  for (TableRow row : table.rows()){
    particles[i] = new PVector(float(row.getString(4)) * width, float(row.getString(5)) * height, float(row.getString(6))*width);
    titles[i] = row.getString(0);
    classes[i] = int(row.getString(1)); // important for numClasses calculation
    errors[i] = float(row.getString(2));
    publishingYears[i] = int(row.getString(3)); // important for min / max year calculation
    
    books[i] = new Book(
      row.getString(0), // title
      int(row.getString(1)), // topic ID
      float(row.getString(2)), // error
      int(row.getString(3)), // publishing year
      new PVector(float(row.getString(4)) * width, float(row.getString(5)) * height, float(row.getString(6))*width) // vector in 3D space
    );
    
    i += 1;
  }
  
  numClasses = max(classes);
  
  // associate books to topics
  topics = new Topic[numClasses];
  for (int j = 0; j < numClasses; j++){
    topics[j] = new Topic(j);
    for (Book book : books){
      if (book.topicID == j){
        topics[j].addBook(book);
      }
    }
  }
  
  // sort all the books in each class
  for (int j = 0; j < numClasses; j++){
    topics[j].sortBooksByError();
  }
  
  // organize everything into years, and generate slices from them
  earliestYear = min(publishingYears);
  latestYear = max(publishingYears);
  rangeYears = latestYear - earliestYear;
  years = new ArrayList<Year>();
  for (int k = 0; k < rangeYears; k++){
    Year year = new Year(k+earliestYear);
    for (Topic topic : topics){
      int y = year.yearDate;
      color c = color(topic.topicColor);
      year.addTopic(new Topic(topic.booksFromYear(y), c));
    }
    if(year.yearsBooks.size() != 0){
      // make the slice for that publishing year
      year.generateSlice();
      years.add(year);
    }
    i++;
  }
  
  generateVisualizationMesh();
  
  smooth();
  
}

void draw() {
  background(background);
  lights();
  cam.lookAt(x, y, z);
  drawAs2DSpur();
}

void generateVisualizationMesh() {
  // create the visualization shape
  theVisualization = createShape(GROUP);
  //theVisualization.beginShape(QUAD_STRIP);
  for(int j = 0; j < years.size() - 1; j++){
    ArrayList<BookParticle> slice1 = years.get(j).bookParticles;
    ArrayList<BookParticle> slice2 = years.get(j+1).bookParticles;
  
    BookParticle lastPairableParticle = slice1.get(0);
    for(int k = 0; k < slice2.size(); k++){
      
      PShape panel = createShape();
      panel.beginShape(QUAD_STRIP);
      
      // FIRST PAIR
      panel.vertex(slice2.get(k).position.x, slice2.get(k).position.y, slice2.get(k).position.z); // first vertex will be in slice2 which will always have more books
      // if the corresponding book is already in the slice before, then connect them
      int idxOfSlice2BPInSlice1 = slice2.get(k).compareTo(slice1);
      if (idxOfSlice2BPInSlice1 >= 0){
        panel.vertex(slice1.get(idxOfSlice2BPInSlice1).position.x, slice1.get(idxOfSlice2BPInSlice1).position.y, slice1.get(idxOfSlice2BPInSlice1).position.z);
        lastPairableParticle = slice1.get(idxOfSlice2BPInSlice1);
      } // otherwise, connect it to the same place that the last slice1 bookParticle did
      else {
        panel.vertex(lastPairableParticle.position.x, lastPairableParticle.position.y, lastPairableParticle.position.z);
      }
      
      // SECOND PAIR (the modulo means that if we go past the size of the slice, we connect to the book pair that we initially started with)
      panel.vertex(slice2.get((k+1) % slice2.size()).position.x, slice2.get((k+1) % slice2.size()).position.y, slice2.get((k+1) % slice2.size()).position.z); // first vertex will be in slice2 which will always have more books
      // if the corresponding book is already in the slice before, then connect them
      int idx2 = slice2.get((k+1) % slice2.size()).compareTo(slice1); // slice2.get((k+1) % slice2.size())
      if (idx2 >= 0){
        panel.vertex(slice1.get(idx2).position.x, slice1.get(idx2).position.y, slice1.get(idx2).position.z);
        lastPairableParticle = slice1.get(idx2);
      } // otherwise, connect it to the same place that the last slice1 bookParticle did
      else {
        panel.vertex(lastPairableParticle.position.x, lastPairableParticle.position.y, lastPairableParticle.position.z);
      }
      panel.endShape();
      
      // set the panel color
      panel.setFill(slice2.get(k).c);
      
      theVisualization.addChild(panel);
    }
  }
}
